/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
/*
** GLW_IMP.C
**
** This file contains ALL Win32 specific stuff having to do with the
** OpenGL refresh.  When a port is being made the following functions
** must be implemented by the port:
**
** GLimp_EndFrame
** GLimp_Init
** GLimp_Shutdown
**
*/
#include <assert.h>
#include <windows.h>
#include "d3d_local.h"
#include "d3dw_win.h"
#include "winquake.h"
#include "resource.h"

IDirect3D9 *d3d_Object = NULL;
IDirect3DDevice9 *d3d_Device = NULL;

D3DCAPS9 d3d_DeviceCaps;
qboolean d3d_DeviceLost = false;
D3DPRESENT_PARAMETERS d3d_PresentParameters;
D3DDISPLAYMODE d3d_DesktopMode;

CRenderToTexture rtt_Brightpass;

cvar_t *vid_lastwindowedmode;
cvar_t *vid_lastfullscreenmode;

struct vidhandler_t
{
	d3d9func_t LoadObjects;
	d3d9func_t ShutDown;
	d3d9func_t LostDevice;
	d3d9func_t ResetDevice;
};


#define MAX_VID_HANDLERS 128

static vidhandler_t vid_handlers[MAX_VID_HANDLERS];
static int numvidhandlers = 0;


CD3DHandler::CD3DHandler (d3d9func_t loadobjects, d3d9func_t shutdown, d3d9func_t lostdevice, d3d9func_t resetdevice)
{
	if (numvidhandlers >= MAX_VID_HANDLERS)
	{
		// this can't throw a regular error as it happens during static initialization time
		return;
	}

	vid_handlers[numvidhandlers].LoadObjects = loadobjects;
	vid_handlers[numvidhandlers].ShutDown = shutdown;
	vid_handlers[numvidhandlers].LostDevice = lostdevice;
	vid_handlers[numvidhandlers].ResetDevice = resetdevice;

	numvidhandlers++;
}


void VID_RunHandlers (int handlerflags)
{
	if (!d3d_Device) return;

	for (int i = 0; i < numvidhandlers; i++)
	{
		// destroy handlers are run before create handlers (in case these are ever used to destroy then immediately recreate objects)
		if ((handlerflags & VH_SHUTDOWN) && vid_handlers[i].ShutDown) vid_handlers[i].ShutDown ();
		if ((handlerflags & VH_LOSTDEVICE) && vid_handlers[i].LostDevice) vid_handlers[i].LostDevice ();
		if ((handlerflags & VH_LOADOBJECTS) && vid_handlers[i].LoadObjects) vid_handlers[i].LoadObjects ();
		if ((handlerflags & VH_RESETDEVICE) && vid_handlers[i].ResetDevice) vid_handlers[i].ResetDevice ();
	}
}


void D3D_GetPresentParameters (void)
{
	IDirect3DSwapChain9 *sc = NULL;

	if (SUCCEEDED (d3d_Device->GetSwapChain (0, &sc)))
	{
		sc->GetPresentParameters (&d3d_PresentParameters);
		sc->Release ();
	}
	else VID_Error (ERR_FATAL, "D3D_GetPresentParameters: could not get present parameters for the swapchain!");
}


D3DPRESENT_PARAMETERS *D3D_GetPresentParameters (D3DDISPLAYMODE *mode)
{
	static D3DPRESENT_PARAMETERS d3d_PP;

	memset (&d3d_PP, 0, sizeof (D3DPRESENT_PARAMETERS));

	if (!mode)
	{
		d3d_PP.Windowed = TRUE;
		d3d_PP.SwapEffect = D3DSWAPEFFECT_DISCARD;
	}
	else
	{
		if (vid_fullscreen->value)
		{
			d3d_PP.Windowed = FALSE;
			d3d_PP.BackBufferFormat = mode->Format;
			d3d_PP.FullScreen_RefreshRateInHz = mode->RefreshRate;
		}
		else
		{
			d3d_PP.Windowed = TRUE;
			d3d_PP.BackBufferFormat = D3DFMT_UNKNOWN;
			d3d_PP.FullScreen_RefreshRateInHz = 0;
		}

		d3d_PP.EnableAutoDepthStencil = TRUE;
		d3d_PP.AutoDepthStencilFormat = D3DFMT_D24S8;
		d3d_PP.Flags = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
		d3d_PP.SwapEffect = D3DSWAPEFFECT_DISCARD;
		d3d_PP.BackBufferCount = 3;
		d3d_PP.BackBufferWidth = mode->Width;
		d3d_PP.BackBufferHeight = mode->Height;
		d3d_PP.hDeviceWindow = glw_state.Window;
		d3d_PP.MultiSampleQuality = 0;
		d3d_PP.MultiSampleType = D3DMULTISAMPLE_NONE;

		if (vid_vsync->value)
			d3d_PP.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
		else d3d_PP.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	}

	return &d3d_PP;
}


std::vector<D3DDISPLAYMODE> d3d_ModeList;

void D3D_InitModeList (void)
{
	// see if the main object is available and if not create a new it
	if (!d3d_Object)
	{
		if (!(d3d_Object = Direct3DCreate9 (D3D_SDK_VERSION)))
		{
			VID_Error (ERR_FATAL, "D3D_InitModeList: Direct3DCreate9 failed\n");
			return;
		}
	}

	// get the desktop mode for reference
	d3d_Object->GetAdapterDisplayMode (D3DADAPTER_DEFAULT, &d3d_DesktopMode);

	// get the count of modes for this format
	int ModeCount = d3d_Object->GetAdapterModeCount (D3DADAPTER_DEFAULT, D3DFMT_X8R8G8B8);

	// no modes available for this format
	if (!ModeCount)
	{
		VID_Error (ERR_FATAL, "D3D_InitModeList: Did not enumerate any modes\n");
		return;
	}

	// enumerate them all
	for (int i = 0; i < ModeCount; i++)
	{
		D3DDISPLAYMODE mode;

		// get the mode description
		if (FAILED (d3d_Object->EnumAdapterModes (D3DADAPTER_DEFAULT, D3DFMT_X8R8G8B8, i, &mode))) break;

		// don't allow modes < 640 x 480
		if (mode.Width < 640) continue;
		if (mode.Height < 480) continue;

		// if the mode width is < height we assume that we have a monitor capable of rotating it's desktop
		// and therefore we skip the mode
		if (mode.Width < mode.Height) continue;

		// add a new mode
		d3d_ModeList.push_back (mode);
	}

	// ensure that we got some modes
	if (!d3d_ModeList.size ()) VID_Error (ERR_FATAL, "D3D_InitModeList: Did not enumerate any modes\n");

	// store out modes for alt-enter handling
	if (!vid_lastwindowedmode) vid_lastwindowedmode = Cvar_Get ("vid_lastwindowedmode", "0", CVAR_ARCHIVE);
	if (!vid_lastfullscreenmode) vid_lastfullscreenmode = Cvar_Get ("vid_lastfullscreenmode", va ("%i", d3d_ModeList.size () - 1), CVAR_ARCHIVE);
}


int VID_BoundMode (int modenum)
{
	if (modenum < 0)
		return 0;
	else if (modenum >= d3d_ModeList.size () - 1)
		return d3d_ModeList.size () - 1;
	else return modenum;
}


void D3D_BuildAnisotropicFilterModes (char ***amodelist)
{
	int i;
	int amodes = 0;

	for (i = d3d_DeviceCaps.MaxAnisotropy; i; i >>= 1)
		amodes++;

	if (amodes == 0)
	{
		// no modes found (can this ever happen???)
		amodelist[0] = (char **) malloc (2 * sizeof (char *));
		amodelist[0][0] = (char *) malloc (10);
		strcpy (amodelist[0][0], "off");
		amodelist[0][1] = NULL;
	}
	else
	{
		amodelist[0] = (char **) malloc ((amodes + 1) * sizeof (char *));
		amodelist[0][0] = (char *) malloc (10);
		strcpy (amodelist[0][0], "off");

		int af;

		for (i = 1, af = 2; i < amodes; i++, af <<= 1)
		{
			amodelist[0][i] = (char *) malloc (10);
			sprintf (amodelist[0][i], "%i x AF", af);
			amodelist[0][i + 1] = NULL;
		}
	}
}


void D3D_PopulateVideoMenu (char ***vmodelist, char ***rrlist)
{
	// add 1 for NULL terming
	vmodelist[0] = (char **) malloc ((d3d_ModeList.size () + 1) * sizeof (char *));
	rrlist[0] = (char **) malloc ((d3d_ModeList.size () + 1) * sizeof (char *));

	for (int j = 0; j < d3d_ModeList.size (); j++)
	{
		vmodelist[0][j] = (char *) malloc (64);
		rrlist[0][j] = (char *) malloc (32);

		// to do - just use the mode list and build it at runtime...?  (will Q2's menu system allow that?)
		sprintf (vmodelist[0][j], "%i x %i", d3d_ModeList[j].Width, d3d_ModeList[j].Height);
		sprintf (rrlist[0][j], "%i Hz", d3d_ModeList[j].RefreshRate);

		// ensure that the list is always null-termed by nulling the next
		vmodelist[0][j + 1] = NULL;
		rrlist[0][j + 1] = NULL;
	}
}


glwstate_t glw_state;

static qboolean VerifyDriver (void)
{
	// see if the main object is available and if not create a new it
	if (!d3d_Object)
	{
		if (!(d3d_Object = Direct3DCreate9 (D3D_SDK_VERSION)))
		{
			VID_Error (ERR_FATAL, "D3D_InitModeList: Direct3DCreate9 failed\n");
			return false;
		}
	}

	// get caps
	d3d_Object->GetDeviceCaps (0, D3DDEVTYPE_HAL, &d3d_DeviceCaps);

	int vsvermaj = D3DSHADER_VERSION_MAJOR (d3d_DeviceCaps.VertexShaderVersion);
	int psvermaj = D3DSHADER_VERSION_MAJOR (d3d_DeviceCaps.PixelShaderVersion);

	// D3D9 will never report > 3 but just in case a future version (if there ever is one) does...
	if (vsvermaj < 3 || psvermaj < 3)
	{
		VID_Error (ERR_FATAL, "GLimp_SetMode : unsupported VS version (%i) or PS version (%i).\nDirectQII needs Shader Model 3.0 minimum.", vsvermaj, psvermaj);
		return false;
	}

	// check other selected caps will go here

	// a bunch of check device format calls will go here
	return true;
}


/*
** GLimp_SetMode
*/
int GLimp_SetMode (D3DDISPLAYMODE *vd)
{
	// store out mode params
	memcpy (vd, &d3d_ModeList[VID_BoundMode ((int) vid_mode->value)], sizeof (D3DDISPLAYMODE));

	VID_Printf (PRINT_ALL, "Initializing Direct3D display\n");

	// direct3d doesn't need to do a CDS
	// this is the global work area we'll use for everything and we only need to get it once
	SystemParametersInfo (SPI_GETWORKAREA, 0, &glw_state.WorkArea, 0);

	// Register the frame class
	WNDCLASS wc = {
		CS_HREDRAW | CS_VREDRAW,
		MainWndProc,
		0,
		0,
		GetModuleHandle (NULL),
		NULL,
		LoadCursor (NULL, IDC_ARROW),
		(HBRUSH) (void *) COLOR_GRAYTEXT,
		NULL,
		WINDOW_CLASS_NAME
	};

	if (!RegisterClass (&wc))
		VID_Error (ERR_FATAL, "GLimp_SetMode : Couldn't register window class");

	// the window is initially created invisible and then a device reset is issued to bring on the full mode
	// it must otherwise be created with full styles otherwise we lose aero styling when we switch w->fs->w
	glw_state.Window = CreateWindowEx (
		ExWindowStyle,
		WINDOW_CLASS_NAME,
		"DirectQII",
		WindowStyle & ~WS_VISIBLE,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		NULL,
		NULL,
		GetModuleHandle (NULL),
		NULL
	);

	if (!glw_state.Window)
		VID_Error (ERR_FATAL, "GLimp_SetMode : Couldn't create window");

	if (!VerifyDriver ())
	{
		VID_Error (ERR_FATAL, "GLimp_SetMode: VerifyDriver failed\n");
		goto fail;
	}

	// see if the main object is available and if not create a new it
	if (!d3d_Object)
	{
		if (!(d3d_Object = Direct3DCreate9 (D3D_SDK_VERSION)))
		{
			VID_Error (ERR_FATAL, "GLimp_SetMode : Direct3DCreate9 failed\n");
			goto fail;
		}
	}

	// prefer a hardware VP device
	HRESULT hr;
	DWORD d3d_VPFlag[] = {D3DCREATE_HARDWARE_VERTEXPROCESSING, D3DCREATE_SOFTWARE_VERTEXPROCESSING, 0};

	for (int i = 0; ; i++)
	{
		if (d3d_VPFlag[i] == 0) goto fail;

		hr = d3d_Object->CreateDevice (
			D3DADAPTER_DEFAULT,
			D3DDEVTYPE_HAL,
			glw_state.Window,
			d3d_VPFlag[i] | D3DCREATE_FPU_PRESERVE,
			D3D_GetPresentParameters (NULL),
			&d3d_Device
		);

		if (SUCCEEDED (hr)) break;
	}

	if (FAILED (hr)) goto fail;

	// store the present parameters it was created with
	D3D_GetPresentParameters ();

	// run a device reset here to bring on the proper mode
	VID_ResetDevice ();

	// clear to black
	d3d_Device->Clear (0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB (0, 0, 0), 1.0f, 1);
	d3d_Device->Present (NULL, NULL, NULL, NULL);

	// and bring on everything else
	VID_RunHandlers (VH_LOADOBJECTS);

	// ensure that the initial mode set doesn't trigger a mode change in the first frame
	vid_mode->modified = false;
	vid_fullscreen->modified = false;
	vid_vsync->modified = false;

	// done
	return rserr_ok;

fail:;
	SAFE_RELEASE (d3d_Device);
	SAFE_RELEASE (d3d_Object);

	return rserr_invalid_mode;
}

/*
** GLimp_Shutdown
**
** This routine does all OS specific shutdown procedures for the OpenGL
** subsystem.  Under OpenGL this means NULLing out the current DC and
** HGLRC, deleting the rendering context, and releasing the DC acquired
** for the window.  The state structure is also nulled out.
**
*/
void GLimp_Shutdown (void)
{
	if (d3d_Device)
	{
		// go back to windowed mode for a clean shutdown
		if (vid_fullscreen->value)
		{
			// hack the windowed mode into the cvars so that they won't write out invalid values to the configs
			vid_mode->value = 0;
			vid_fullscreen->value = 0;
			VID_ResetDevice ();
		}

		// clear to black for going down
		d3d_Device->Clear (0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB (0, 0, 0), 1.0f, 1);
		d3d_Device->Present (NULL, NULL, NULL, NULL);
	}

	// D3D HERE
	// also release all textures and other gubbins
	// these are NOT automatically destroyed with the "rendering context"
	GL_ShutdownImages ();
	VID_RunHandlers (VH_SHUTDOWN | VH_LOSTDEVICE);

	SAFE_RELEASE (d3d_Device);
	SAFE_RELEASE (d3d_Object);

	if (glw_state.Window)
	{
		// hide the window before destroying it to remove left-over taskbar icons
		// (i haven't seen this since i moved off win98 so i guess that recent versions do it automatically)
		ShowWindow (glw_state.Window, SW_HIDE);
		DestroyWindow (glw_state.Window);
		glw_state.Window = NULL;
	}

	UnregisterClass (WINDOW_CLASS_NAME, GetModuleHandle (NULL));
}


/*
** GLimp_BeginFrame
*/
qboolean GLimp_TestDevice (void)
{
	if (d3d_DeviceLost)
	{
		// yield some CPU time
		Sleep (5);

		// see if the device can be recovered
		HRESULT hr = d3d_Device->TestCooperativeLevel ();

		switch (hr)
		{
		case D3D_OK:
			// the device is no longer lost
			VID_RunHandlers (VH_RESETDEVICE);
			d3d_DeviceLost = false;
			VID_Printf (PRINT_ALL, "Lost device recovery complete\n");
			break;

		case D3DERR_DEVICELOST:
			// the device cannot be recovered at this time
			break;

		case D3DERR_DEVICENOTRESET:
			// the device is ready to be reset
			// HOWEVER, instead of resetting it, we'll reload the full refresh.  this is a bit cheesy but it
			// ensures that we'll get everything back the way it should be, provides a single code path for
			// it all, and handles resolution changing properly.
			VID_RunHandlers (VH_LOSTDEVICE);
			d3d_Device->Reset (&d3d_PresentParameters);
			break;

		case D3DERR_DRIVERINTERNALERROR:
		default:
			// something bad happened
			// note: this isn't really a proper clean-up path as it doesn't do input/sound/etc cleanup either...
			VID_Error (ERR_FATAL, "GLimp_BeginFrame: D3DERR_DRIVERINTERNALERROR");
			break;
		}

		// no rendering this frame
		return false;
	}

	// device is not lost so render this frame
	return true;
}


void GLimp_BeginFrame (float camera_separation)
{
	d3d_Device->BeginScene ();

	if (vid_gamma->value != 1.0f)
		rtt_Brightpass.Begin (d3d_BrightpassTexture);
}


/*
** GLimp_EndFrame
**
** Responsible for doing a swapbuffers and possibly for other stuff
** as yet to be determined.  Probably better not to make this a GLimp
** function and instead do a call to GLimp_SwapBuffers.
*/
void Draw_FPS (void);

void GLimp_EndFrame (void)
{
	if (d3d_DeviceLost) return;

	Draw_FPS ();

	// finish any pending 2D drawing
	D3D_End2D ();

	// end our rtt passes
	rtt_Brightpass.End (RTTPASS_BRIGHTPASS);

	d3d_Device->EndScene ();

	HRESULT hr = d3d_Device->Present (NULL, NULL, NULL, NULL);

	// check for a lost device
	if (hr == D3DERR_DEVICELOST) d3d_DeviceLost = true;
}


/*
** GLimp_AppActivate
*/
void GLimp_AppActivate (qboolean active)
{
	if (active)
		ShowWindow (glw_state.Window, SW_RESTORE);
	else ShowWindow (glw_state.Window, SW_MINIMIZE);
}


void VID_ResetDevice (void)
{
	// this is the standard device reset that comes from a mode change
	if (!d3d_Device) return;

	// pull out the video mode number but don't change the cvar so that we won't trigger another reset
	int modenum = VID_BoundMode ((int) vid_mode->value);
	HRESULT hr = S_OK;
	qboolean borderless = false;

	D3DDISPLAYMODE *mode = &d3d_ModeList[modenum];

	// borderless modes don't exist in fullscreen
	if (!vid_fullscreen->value)
	{
		if (mode->Width >= (glw_state.WorkArea.right - glw_state.WorkArea.left)) borderless = true;
		if (mode->Height >= (glw_state.WorkArea.bottom - glw_state.WorkArea.top)) borderless = true;
	}

	if (borderless)
	{
		// borderless only makes sense at the full native res
		modenum = d3d_ModeList.size () - 1;
		mode = &d3d_ModeList[modenum];
	}

	// make sure that we're ready to reset
	while ((hr = d3d_Device->TestCooperativeLevel ()) != D3D_OK)
	{
		if (FAILED (hr))
		{
			// at this stage nothing has been destroyed so we can safely error out
			VID_Printf (PRINT_ALL, "VID_ResetDevice: IDirect3DDevice9::TestCooperativeLevel failed\n");
			return;
		}

		Sys_SendKeyEvents ();
		Sleep (1);
	}

	// release default resources
	VID_RunHandlers (VH_LOSTDEVICE);

	// these are needed even when going to a fullscreen mode otherwise we'll lose the aero theme when we go back to windowed
	// and they need to happen before the reset call too otherwise none of this will work the way it's meant to
	SetWindowLong (glw_state.Window, GWL_EXSTYLE, (vid_fullscreen->value || borderless) ? 0 : ExWindowStyle);
	SetWindowLong (glw_state.Window, GWL_STYLE, (vid_fullscreen->value || borderless) ? 0 : WindowStyle);
	SetWindowPos (glw_state.Window, NULL, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER | SWP_FRAMECHANGED);

	// and reset the device
	if (FAILED (d3d_Device->Reset (D3D_GetPresentParameters (mode))))
	{
		VID_RunHandlers (VH_RESETDEVICE);
		VID_Printf (PRINT_ALL, "VID_ResetDevice: IDirect3DDevice9::Reset failed\n");
		return;
	}

	// make sure that the reset has completed
	while ((hr = d3d_Device->TestCooperativeLevel ()) != D3D_OK)
	{
		if (FAILED (hr))
		{
			// at this stage stuff has been destroyed so we must crash out
			VID_Error (ERR_FATAL, "VID_ResetDevice: IDirect3DDevice9::TestCooperativeLevel failed\n");
			return;
		}

		Sys_SendKeyEvents ();
		Sleep (1);
	}

	// update video width and height (done early so that rendertargets will have access to it correctly)
	memcpy (&viddef, mode, sizeof (D3DDISPLAYMODE));

	VID_RunHandlers (VH_RESETDEVICE);

	// store the present parameters it was reset to
	D3D_GetPresentParameters ();

	// update the window rect correctly for all mode types
	SetRect (&glw_state.WindowRect, 0, 0, mode->Width, mode->Height);

	// because we changed the style we need to send a SWP_FRAMECHANGED here - we also put the window to the top of the z-order
	// if a windowed mode we also need to set the new window height and width after the reset call otherwise reset will just
	// overwrite the mode size with it's own former size (gee, this crap should really be documented)
	if (vid_fullscreen->value)
	{
		// remind me again of why i'm doing this...
		SetWindowPos (
			glw_state.Window,
			HWND_TOP,
			0, 0, 0, 0,
			SWP_NOMOVE | SWP_NOSIZE
		);

		Cvar_SetValue ("vid_lastfullscreenmode", modenum);
	}
	else if (borderless)
	{
		SetWindowPos (
			glw_state.Window,
			HWND_TOP,
			(d3d_DesktopMode.Width - (glw_state.WindowRect.right - glw_state.WindowRect.left)) / 2,
			(d3d_DesktopMode.Height - (glw_state.WindowRect.bottom - glw_state.WindowRect.top)) / 2,
			(glw_state.WindowRect.right - glw_state.WindowRect.left),
			(glw_state.WindowRect.bottom - glw_state.WindowRect.top),
			0
		);

		Cvar_SetValue ("vid_lastwindowedmode", modenum);
	}
	else
	{
		// fixme - cache this icon handle
		HICON hIcon = LoadIcon (GetModuleHandle (NULL), MAKEINTRESOURCE (IDI_ICON1));

		SendMessage (glw_state.Window, WM_SETICON, (WPARAM) TRUE, (LPARAM) hIcon);
		SendMessage (glw_state.Window, WM_SETICON, (WPARAM) FALSE, (LPARAM) hIcon);

		AdjustWindowRectEx (&glw_state.WindowRect, WindowStyle, FALSE, ExWindowStyle);

		SetWindowPos (
			glw_state.Window,
			HWND_TOP,
			glw_state.WorkArea.left + ((glw_state.WorkArea.right - glw_state.WorkArea.left) - (glw_state.WindowRect.right - glw_state.WindowRect.left)) / 2,
			glw_state.WorkArea.top + ((glw_state.WorkArea.bottom - glw_state.WorkArea.top) - (glw_state.WindowRect.bottom - glw_state.WindowRect.top)) / 2,
			(glw_state.WindowRect.right - glw_state.WindowRect.left),
			(glw_state.WindowRect.bottom - glw_state.WindowRect.top),
			0
		);

		Cvar_SetValue ("vid_lastwindowedmode", modenum);
	}

	// update video width and height
	memcpy (&viddef, mode, sizeof (D3DDISPLAYMODE));

	VID_Printf (PRINT_ALL, "...setting mode %ix%i %s\n", mode->Width, mode->Height, (vid_fullscreen->value ? "fullscreen" : "windowed"));

	// let the sound and input subsystems know about the new window
	VID_NewWindow (mode->Width, mode->Height);

	// default states are lost so restore them
	GL_SetDefaultState ();

	// ensure that the window becomes active
	GLimp_AppActivate (true);
}


void VID_HandleAltEnter (void)
{
	if (vid_fullscreen->value)
	{
		Cvar_SetValue ("vid_fullscreen", 0);
		Cvar_SetValue ("vid_mode", vid_lastwindowedmode->value);
	}
	else
	{
		Cvar_SetValue ("vid_fullscreen", 1);
		Cvar_SetValue ("vid_mode", vid_lastfullscreenmode->value);
	}
}

